#include "tyaconfig.h"
#include <stdio.h>
#include <native.h>
#include <monitor.h>

#ifdef JDK12
#include <jit.h>
#endif

#include "tya.h"

#define TRUE 1
#define FALSE 0

// Maximum size of the variable
//assignement generated code
#define MAX_CODE_LENGTH 4096
#define MAX_STACK_SIZE 1024
#define MAX_LOCAL_VAR 256

// A structure for the optimization algorithm,
// which stores a to-be generated code
typedef struct variableDetermination
{
	// Local variable information
	char valIsKnown ;

	// Float so it can adapt with all types
	float value ;

	// The code which made this variable this value
	short code_length ;
	char *code ;
} variableDetermination ;

// An array of local variables
variableDetermination localVar[MAX_LOCAL_VAR] ;

// A small stack for the simulation
variableDetermination stack[MAX_STACK_SIZE] ;
	
// Instruction table structure
typedef struct instrOffset
{
	int offset ;
	char begBlock ;
} instrOffset ;

// Instructions index table
instrOffset *instrTable ;

// Number of instruction
int instrNumber ;

// Method 32-bit alignment
int alignment32 ;

// Usefull function signatures
int readU2(unsigned char *p) ;
int readU4(unsigned char *p) ;

// Number of padding bytes for lookup- and table- switch instructions
int paddingBytes (unsigned char *code)
{
	int i = 0 ;
	while (((int)code++ % 4 != alignment32) && (i<4))
	{
		if (*code != 0)
		{
			fprintf (stderr, "Bytecode syntax error !\n") ;
			exit (1) ;
		}
		else
			i++ ;
	}
	return i ;
}

// Search an instruction index by its offset
// Return instruction index in the table
// The algorithm is dichotomic
int searchInstr (int offset)
{
	int i, half ;
	half = instrNumber/2 ;
	i = half ;
	while (instrTable[i].offset != offset)
	{
		half = half/2 ? half/2 : 1 ;
		i += instrTable[i].offset < offset ? half : -half ;
	}
	return i ;
}

// Create BB table
// return -1 if error, number of blocks otherwise
int basicBlockDetection (struct CINFO *cinfo)
{
	int i, j, n ;

	// Looking for sequence rupture instructions
	i = n = 0 ;
	while (i < instrNumber)
	{
		// We are at the begining of a basic block
		n++ ;
		instrTable[i].begBlock = TRUE ;

		// For each instruction opcodes, if one correspond to a
		// sequence rupture instruction, then this is the last
		// instruction of the block, and the next one is the
		// begining of a new block
		while (i < instrNumber &&
				(*(cinfo->bptr+instrTable[i].offset) < 0x99 ||
				(*(cinfo->bptr+instrTable[i].offset) > 0xB1 && *(cinfo->bptr+instrTable[i].offset) < 0xB6) ||
				(*(cinfo->bptr+instrTable[i].offset) > 0xB9 && *(cinfo->bptr+instrTable[i].offset) < 0xBF) ||
				(*(cinfo->bptr+instrTable[i].offset) > 0xBF && *(cinfo->bptr+instrTable[i].offset) < 0xC6) ||
				(*(cinfo->bptr+instrTable[i].offset) > 0xC9 &&
				*(cinfo->bptr+instrTable[i].offset) != 0xD7 &&
				*(cinfo->bptr+instrTable[i].offset) != 0xD9 &&
				*(cinfo->bptr+instrTable[i].offset) != 0xDA &&
				*(cinfo->bptr+instrTable[i].offset) != 0xE2 &&
				*(cinfo->bptr+instrTable[i].offset) != 0xE9)
				))
			i++ ;

		// Among these instructions, there are some which are
		// jumps, with an offset pointing to a block-begining.
		
		// There are opcodes with a 16 bit offset.
		if ((*(cinfo->bptr+instrTable[i].offset) >= 0x99 &&
					*(cinfo->bptr+instrTable[i].offset) <= 0xA8) ||
					*(cinfo->bptr+instrTable[i].offset) == 0xC6 ||
					*(cinfo->bptr+instrTable[i].offset) == 0xC7)
		{
#ifdef VERBOSE
			lprintf ("Index %d : offset %d -> index %d\n",
					i,
					(signed short)readU2 (cinfo->bptr+instrTable[i].offset+1),
					searchInstr (instrTable[i].offset+(signed short)readU2 (cinfo->bptr+instrTable[i].offset+1))) ;
#endif
			// Search the instruction where we want to jump, and
			// mark it as a block beginning
			instrTable [searchInstr (instrTable[i].offset+(signed short)readU2 (cinfo->bptr+instrTable[i].offset+1))].begBlock = TRUE ;
		} // 16 bits

		// And some with 32 bit offset.
		if (*(cinfo->bptr+instrTable[i].offset) == 0xC8 || *(cinfo->bptr+instrTable[i].offset) == 0xC9)
		{
#ifdef VERBOSE
			lprintf ("Index %d : offset %d -> index %d\n",
					i,
					(signed int)readU4 (cinfo->bptr+instrTable[i].offset+1),
					searchInstr (instrTable[i].offset+(signed int)readU4 (cinfo->bptr+instrTable[i].offset+1))) ;
#endif
			// Search the instruction where we want to jump, and
			// mark it as a block beginning
			instrTable [searchInstr (instrTable[i].offset+(signed int)readU4 (cinfo->bptr+instrTable[i].offset+1))].begBlock = TRUE ;
		} // 32 bits

		// tableswitch & lookupswitch
		if (*(cinfo->bptr+instrTable[i].offset) == 0xAA || *(cinfo->bptr+instrTable[i].offset) == 0xAB)
		{
			int tmpOffset, low, high, npairs ;

			tmpOffset = instrTable[i].offset+1 ;

			// Null padding bytes
			tmpOffset += paddingBytes (cinfo->bptr+tmpOffset) ;

			// Default
#ifdef VERBOSE
			lprintf ("Index %d : offset %d -> index %d\n",
					i,
					readU4 (cinfo->bptr+tmpOffset),
					searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))) ;
#endif
			// Search the instruction where we want to jump, and
			// mark it as a block beginning
			instrTable [searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))].begBlock = TRUE ;

			// Tableswitch
			if (*(cinfo->bptr+instrTable[i].offset) == 0xAA)
			{
				// Indexed offsets
				tmpOffset += 4 ; low = readU4 (cinfo->bptr+tmpOffset) ;
				tmpOffset += 4 ; high = readU4 (cinfo->bptr+tmpOffset) ;
				for (j=0; j<high-low+1; j++)
				{
					tmpOffset += 4 ;
#ifdef VERBOSE
					lprintf ("Index %d : offset %d -> index %d\n",
							i,
							readU4 (cinfo->bptr+tmpOffset),
							searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))) ;
#endif
					// Search the instruction where we want to jump, and
					// mark it as a block beginning
					instrTable [searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))].begBlock = TRUE ;
				}
			}
			// Lookupswitch
			else
			{
				// Matching offsets
				tmpOffset += 4 ; npairs = readU4 (cinfo->bptr+tmpOffset) ;
				for (j=0; j<npairs; j++)
				{
					tmpOffset += 8 ;
#ifdef VERBOSE
					lprintf ("Index %d : offset %d -> index %d\n",
							i,
							readU4 (cinfo->bptr+tmpOffset),
							searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))) ;
#endif
					// Search the instruction where we want to jump, and
					// mark it as a block beginning
					instrTable [searchInstr (instrTable[i].offset+readU4 (cinfo->bptr+tmpOffset))].begBlock = TRUE ;
				}
			}
		}

		i++ ;
	}

#ifdef VERBOSE
	// Debugging information
	for (i=0; i<instrNumber; i++)
	{
		lprintf ("\nIndex %d *** Opcode=%02x, ",
				i,
				*(cinfo->bptr+instrTable[i].offset)) ;
		j = instrTable[i].offset+1 ;
		while (j < instrTable[i+1].offset && j < instrTable[instrNumber-1].offset)
			lprintf ("%02x ", *(cinfo->bptr+j++)) ;
		lprintf ("- Begin block : %s\n", instrTable[i].begBlock ? "Yes" : "") ;
	}
#endif

	return (n) ;
}

// Create instructions index table.
// returns -1 if error, number of instructions otherwise
int indexInstructions (struct CINFO *cinfo, int length)
{
	// counters
	int i, n ;

	// All instructions
	instrTable = (instrOffset *) sysMalloc (length*sizeof(instrOffset)) ;

	// Switch tables must be aligned, but we don't
	// know how
	alignment32 = (int)cinfo->bptr % 4 ;

	i = n = 0 ;
	// Decoding instructions
	while (i < length)
	{
		instrTable[n].offset = i ;
		instrTable[n++].begBlock = FALSE ;
		i += decode_bytecode (cinfo->bptr+i) ;
	}

	instrNumber = n ;
	return (n) ;
}

// Generate the code which restore the state of the stack.
int flushStack (char *newCode, int newCode_Length, int stack_top)
{
	int j, k ;

	for (j=0; j<stack_top+1; j++)
	{
		for (k=0; k<stack[j].code_length; k++)
		{
#ifdef VERBOSE2
			lprintf ("%02x ", stack[j].code[k]) ;
#endif
			newCode[newCode_Length++] = stack[j].code[k] ;
		}
		stack[j].code_length = 0 ;
#ifdef VERBOSE2
		lprintf ("\n") ;
#endif
	}
	return newCode_Length ;
}

// Generate the code which restore variables values
int flushVars (char *newCode, int newCode_Length)
{
	int j, k ;

	for (j=0; j<MAX_LOCAL_VAR; j++)
	{
		if (localVar[j].code_length > 0)
		{
			for (k=0; k<localVar[j].code_length; k++)
			{
#ifdef VERBOSE2
				lprintf ("%02x ", localVar[j].code[k]) ;
#endif
				newCode[newCode_Length++] = localVar[j].code[k] ;
			}
#ifdef VERBOSE2
			lprintf ("\n") ;
#endif
			localVar[j].code_length = 0 ;
		}
	}
	return newCode_Length ;
}

// ***************************************************************
// *** T H E   O P T I M I Z A T I O N   M A I N   M E T H O D ***
// ***************************************************************
int tyaoptimize (struct CINFO *cinfo, int code_length)
{
	// The number of basic blocks
	int basicBlockNumber ;

	// The code we will generate after optimizations
	char *newCode ;

	// Some usefull integers
	int i, j, n, k, opcode, currentInstr ;

	// The length of the new code
	int newCode_Length = 0 ;

	// The actual number of elements in the stack
	// (It's empty at the beginning)
	int stack_top = -1 ;

	// The new code is at most as long as the actual one
	newCode = (char *) sysMalloc (code_length*sizeof(char)) ;

	// The length of the stack and variable ones is at most
	// MAX_CODE_LENGTH.
	for (i=0; i<MAX_STACK_SIZE; i++)
	{
		stack[i].valIsKnown = FALSE ;
		stack[i].code_length = 0 ;
		stack[i].code = (char *) sysMalloc (MAX_CODE_LENGTH*sizeof(char)) ;
	}

	// Indexing byte code instructions
	indexInstructions (cinfo, code_length) ;

	// Building basic block structure
	basicBlockNumber = basicBlockDetection (cinfo) ;

	// Local variables initialization to 'unknown value'
	// and without any code making an assignment.
	for (i=0; i<MAX_LOCAL_VAR; i++)
	{
		localVar[i].valIsKnown = FALSE ;
		localVar[i].code_length = 0 ;
		localVar[i].code = (char *) sysMalloc (MAX_CODE_LENGTH*sizeof(char)) ;
	}

	k = n = 0 ;
	currentInstr = 0 ;

	// Now, the effective optimization begins
	while (currentInstr < instrNumber)
	{
		// Here we are at the begining of a block, so we
		// must align our new code as the old one is.
		// To do so, we fill it with nop
		while (newCode_Length < instrTable[currentInstr].offset)
			newCode[newCode_Length++] = 0x00 ;

		// Then, we can examine the block
		while (currentInstr < instrNumber && instrTable[currentInstr].begBlock == FALSE)
		{
			// If the next instruction begins a block,
			// then we must restore the stack and variables to
			// their values, before writing this one, because
			// it can be a jump and its offset has to be kept
			// right.
			if (currentInstr+1 < instrNumber && instrTable[currentInstr+1].begBlock == TRUE)
			{
				// Restoring stack and variables
				newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
				newCode_Length = flushVars (newCode, newCode_Length) ;

				// Our stack is now empty
				stack_top = -1 ;

				// Since the last instruction of the block could be a
				// jump, its offset should be true so it has to
				// be aligned.
				while (newCode_Length < instrTable[currentInstr].offset)
					newCode[newCode_Length++] = 0x00 ;
			}

#ifdef VERBOSE
			lprintf ("currentInstr = %d / %d : offsets : old = %d, new = %d\n", currentInstr, instrNumber, instrTable[currentInstr].offset, newCode_Length) ;
#endif

			opcode = *(cinfo->bptr+instrTable[currentInstr].offset) ;
			// Looking at the current bytecode.
			switch (opcode)
			{
// ******************************************************************************
// ******* iconst *** iconst *** iconst *** iconst *** iconst *** iconst ********
// ******************************************************************************
				// Push the value onto the stack
				case 0x02: // iconst_m1
				case 0x03: // iconst_0
				case 0x04: // iconst_1
				case 0x05: // iconst_2
				case 0x06: // iconst_3
				case 0x07: // iconst_4
				case 0x08: // iconst_5
#ifdef VERBOSE2
					lprintf ("iconst_%d\n", (int)(opcode-3)) ;
#endif
					stack_top++ ;
					stack[stack_top].valIsKnown = TRUE ;
					stack[stack_top].value = (int)(opcode-3) ;
					stack[stack_top].code_length = 0 ;
					stack[stack_top].code[stack[stack_top].code_length++] = opcode ;
					break ;

// ******************************************************************************
// ******* bipush *** bipush *** bipush *** bipush *** bipush *** bipush ********
// ******************************************************************************
				// Push the 8-bit value onto the stack
				case 0x10:
					n = (char)(*(cinfo->bptr+instrTable[currentInstr].offset+1)) ;
#ifdef VERBOSE2
					lprintf ("bipush %d\n", n) ;
#endif
					stack_top++ ;
					stack[stack_top].valIsKnown = TRUE ;
					stack[stack_top].value = (int)n ;
					stack[stack_top].code_length = 0 ;
					stack[stack_top].code[stack[stack_top].code_length++] = opcode ;
					stack[stack_top].code[stack[stack_top].code_length++] = n ;
					break ;

// ******************************************************************************
// **** iload n *** iload n *** iload n *** iload n *** iload n *** iload n *****
// ******************************************************************************
				// Push a variable on the stack:
				// if the value is known, the push this value
				// else just load the variable.
				case 0x15:
					n = (unsigned char)(*(cinfo->bptr+instrTable[currentInstr].offset+1)) ;

				case 0x1a:
				case 0x1b:
				case 0x1c:
				case 0x1d:
					if (opcode == 0x1a) n = 0 ;
					if (opcode == 0x1b) n = 1 ;
					if (opcode == 0x1c) n = 2 ;
					if (opcode == 0x1d) n = 3 ;

					stack_top++ ;
					stack[stack_top].code_length = 0 ;
#ifdef VERBOSE2
					lprintf ("iload %d\n", n) ;
#endif

					// If the variable value is known, then we push
					// its value onto the stack, and the code to
					// produce is the one which push this value onto
					// the jvm stack.
					// To avoid segfault, we must have a new code smaller
					// than the original, tus we test wether the 'load' is
					// longer or not than the 'push'.
					if (localVar[n].valIsKnown && (localVar[n].value >= -32768 && localVar[n].value <= 32767))
					{
						stack[stack_top].valIsKnown = TRUE ;
						stack[stack_top].value = (int)localVar[n].value ;

						// iconst_<i> if i is between -1 and 5
						if (localVar[n].value >= -1 && localVar[n].value <= 5)
							stack[stack_top].code[stack[stack_top].code_length++] = (unsigned char)(0x03+localVar[n].value) ;

						else
						{
							// bipush i if i can be in a signed-byte and n > 3
							if (localVar[n].value >= -128 && localVar[n].value <= 127 && n > 3)
							{
								stack[stack_top].code[stack[stack_top].code_length++] = (unsigned char)0x10 ;
								stack[stack_top].code[stack[stack_top].code_length++] = (signed char)(localVar[n].value) ;
							}

							else
							{
								if (localVar[n].code_length > 0)
									for (j=0; j<localVar[n].code_length; j++)
									{
#ifdef VERBOSE2
										lprintf ("%02x ", localVar[n].code[j]) ;
#endif
										newCode[newCode_Length++] = localVar[n].code[j] ;
									}
#ifdef VERBOSE2
								if (localVar[n].code_length > 0)
									lprintf ("\n") ;
#endif
								localVar[n].code_length = 0 ;

								stack[stack_top].code[stack[stack_top].code_length++] = opcode ;
								if (opcode == 0x15)
									stack[stack_top].code[stack[stack_top].code_length++] = n ;
							}
						}
					}

					// If the value is not known, then we must assign the 
					// variable its value and restore the stack state.
					else
					{
						// flushing stack
						newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
						// Our stack is now empty and will contain the load
						stack_top = 0 ;
						stack[stack_top].valIsKnown = FALSE ;
						if (localVar[n].code_length > 0)
							for (j=0; j<localVar[n].code_length; j++)
							{
#ifdef VERBOSE2
								lprintf ("%02x ", localVar[n].code[j]) ;
#endif
								newCode[newCode_Length++] = localVar[n].code[j] ;
							}
#ifdef VERBOSE2
						if (localVar[n].code_length > 0)
							lprintf ("\n") ;
#endif
						localVar[n].code_length = 0 ;

						// Now, just do the load
						stack[stack_top].code[stack[stack_top].code_length++] = opcode ;
						if (opcode == 0x15)
							stack[stack_top].code[stack[stack_top].code_length++] = n ;
					}
#ifdef VERBOSE2
					if (stack[stack_top].valIsKnown == TRUE)
						lprintf (" value = %d\n", (int)stack[stack_top].value) ;
					else
						lprintf (" unknown value...\n") ;
#endif
					break ;

// ******************************************************************************
// ******** istore n *** istore n *** istore n *** istore n *** istore n ********
// ******************************************************************************
				// Since the bytecode has been comppiled, we could expect
				// that it doesn't contain any error. For instance, if
				// a store instruction occurs, the stack contains at least
				// one element, which is of the right type.
				case 0x36:
					n = (unsigned char)(*(cinfo->bptr+instrTable[currentInstr].offset+1)) ;

				case 0x3b:
				case 0x3c:
				case 0x3d:
				case 0x3e:
					if (opcode == 0x3b) n = 0 ;
					if (opcode == 0x3c) n = 1 ;
					if (opcode == 0x3d) n = 2 ;
					if (opcode == 0x3e) n = 3 ;

					localVar[n].code_length = 0 ;
					localVar[n].valIsKnown = FALSE ;

					// When a store instruction occurs, then the coresponding variable
					// takes the same value as the stack top, including the code which
					// generate the value, if we caught it.
					// If we did not, then this variable have to get the actual stack
					// top, so we must generate the store instruction.
					// Furthermore, if the previous instruction was a load or
					// a push, then it has to be generated to. We must then
					// generate all the stack.
#ifdef VERBOSE2
					lprintf ("istore %d\n", n) ;
#endif
					if (stack_top > -1)
					{
						localVar[n].valIsKnown = stack[stack_top].valIsKnown ;
						if (stack[stack_top].valIsKnown)
							localVar[n].value = (int)stack[stack_top].value ;

						for (j=0; j<stack[stack_top].code_length; j++)
							localVar[n].code[localVar[n].code_length++] = stack[stack_top].code[j] ;

						// Then, the store instruction has to be generated for the variable
						localVar[n].code[localVar[n].code_length++] = opcode ;
						if (opcode == 0x36)
							localVar[n].code[localVar[n].code_length++] = (unsigned char)n ;

#ifdef VERBOSE2
						if (stack[stack_top].valIsKnown)
							lprintf (" value = %d.", (int)stack[stack_top].value) ;
						else
							lprintf (" unknown value...") ;
#endif
						stack[stack_top].code_length = 0 ;
						stack_top-- ;
					}
					else
					{
						// flushing stack
						newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
						// Our stack is now empty
						stack_top = -1 ;
#ifdef VERBOSE2
						lprintf ("%02x ", opcode) ;
#endif
						newCode[newCode_Length++] = opcode ;
						if (opcode == 0x36)
						{
							newCode[newCode_Length++] = (unsigned char)n ;
#ifdef VERBOSE2
							lprintf ("%02x", n) ;
#endif
						}
						localVar[n].code_length = 0 ;
						localVar[n].valIsKnown = FALSE ;
					}
#ifdef VERBOSE2
					lprintf ("\n") ;
#endif
					break ;

// ******************************************************************************
// ***** iadd *** isub *** imul *** idiv *** iadd *** isub *** imul *** idiv ****
// ******************************************************************************
				// Arithmetic operation: the stack top code is
				// the concatenation of the two levels and the
				// operation, or just a push if the two values are known.
				case 0x60: // iadd
				case 0x64: // isub
				case 0x68: // imul
				case 0x6C: // idiv
				case 0x70: // irem
					if (stack_top > 0)
					{
						int result = 0 ;
						int resultIsKnown = stack[stack_top].valIsKnown & stack[stack_top-1].valIsKnown ;
						stack_top-- ;
						if (resultIsKnown)
							switch (opcode)
							{
								// add
								case 0x60: result = stack[stack_top].value+stack[stack_top+1].value ; break ;
								// sub
								case 0x64: result = stack[stack_top].value-stack[stack_top+1].value ; break ;
								// mul
								case 0x68: result = stack[stack_top].value*stack[stack_top+1].value ; break ;
								// div & rem
								// If the divisor is nul, we let the JVM
								// throws an exception.
								case 0x6C:
								case 0x70:
								   if (stack[stack_top+1].value != 0)
									   result = (opcode == 0x6C ? (int)stack[stack_top].value/(int)stack[stack_top+1].value : (int)stack[stack_top].value%(int)stack[stack_top+1].value) ;
								   else
									   resultIsKnown = FALSE ;
							}
						if (resultIsKnown && result >= -32768 && result <= 32767)
						{
							stack[stack_top].valIsKnown = TRUE ;
							stack[stack_top].value = (int)result ;
							stack[stack_top].code_length = 0 ;
#ifdef VERBOSE2
							lprintf ("Arithmetic operation, result = %d\n",
								(int)stack[stack_top].value) ;
#endif

							// iconst_<i> if i is between  and 5
							if (stack[stack_top].value >= -1 && stack[stack_top].value <= 5)
								stack[stack_top].code[stack[stack_top].code_length++] = (unsigned char)(0x03+stack[stack_top].value) ;

							else
							{
								// bipush i if i can be in a signed-byte
								if (stack[stack_top].value >= -128 && stack[stack_top].value <= 127)
								{
									stack[stack_top].code[stack[stack_top].code_length++] = (unsigned char)0x10 ;
									stack[stack_top].code[stack[stack_top].code_length++] = (signed char)(stack[stack_top].value) ;
								}

								// sipush i if i can be in a signed short
								else
								{
									stack[stack_top].code[stack[stack_top].code_length++] = (unsigned char)0x11 ;
									stack[stack_top].code[stack[stack_top].code_length++] = (signed char)((int)(stack[stack_top].value) >> 8) ;
									stack[stack_top].code[stack[stack_top].code_length++] = (signed char)((int)(stack[stack_top].value) & 0xFF) ;
								}
							}
						}
						else
						{
							for (j=0; j<stack[stack_top+1].code_length; j++)
								stack[stack_top].code[stack[stack_top].code_length++] = stack[stack_top+1].code[j] ;
							stack[stack_top].code[stack[stack_top].code_length++] = opcode ;
							stack[stack_top].valIsKnown = resultIsKnown ;
							stack[stack_top].value = resultIsKnown ? result : 0 ;
#ifdef VERBOSE2
							lprintf ("Arith op, result unknown or too big\n") ;
#endif
						}
					}
					else
					{
						newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;

						// Our stack is now empty
						stack_top = -1 ;

						newCode[newCode_Length++] = (unsigned char)opcode ;
#ifdef VERBOSE2
						lprintf ("%02x\n", opcode) ;
#endif
					}
					break ;

// ******************************************************************************
// ***** default ****** default ****** default ****** default ****** default ****
// ******************************************************************************
					// The default case: flush stack and variables, then
					// copy the code
				default:
					// Restoring stack and variables
					newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
					newCode_Length = flushVars (newCode, newCode_Length) ;

					// Our stack is now empty
					stack_top = -1 ;

					for (j=instrTable[currentInstr].offset; j<((currentInstr==instrNumber-1) ? code_length : instrTable[currentInstr+1].offset); j++)
					{
#ifdef VERBOSE2
						lprintf ("%02x ", *(cinfo->bptr+j)) ;
#endif
						newCode [newCode_Length++] = *(cinfo->bptr+j) ;
					}
#ifdef VERBOSE2
					lprintf ("\n") ;
#endif
			}
			currentInstr++ ;
		}

		// End of block
		// Restoring stack and variables
		newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
		newCode_Length = flushVars (newCode, newCode_Length) ;

		// Our stack is now empty
		stack_top = -1 ;

		// All variables are now unknow
		for (i=0; i<MAX_LOCAL_VAR; i++)
			localVar[i].valIsKnown = FALSE ;

#ifdef VERBOSE
		lprintf ("\t\tEnd of block\n") ;
#endif

		instrTable[currentInstr].begBlock = FALSE ;
	}

	// And there is a new need of flush at the end of the method,
	// but we don't care anymore from the stack size or code
	// lengths since the function will return.
	newCode_Length = flushStack (newCode, newCode_Length, stack_top) ;
	newCode_Length = flushVars (newCode, newCode_Length) ;

	// We must copy our code into the old memory block...
	for (i=0; i<newCode_Length; i++)
		*(cinfo->bptr+i) = newCode[i] ;

	// And free memory we used
	sysFree (instrTable) ;
	sysFree (newCode) ;

	for (i=0; i<MAX_STACK_SIZE; i++)
		sysFree (stack[i].code) ;

	for (i=0; i<MAX_LOCAL_VAR; i++)
		sysFree (localVar[i].code) ;
	
	return (code_length) ;
}
